/* xscreensaver freeframe plugin wrapper, Copyright (c) 2004 Gabor Papp
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 */

/* This file contains code to make use of xscreensaver hacks as FreeFrame plugins.
    - include screenhackff.h instead of screenhack.h
    
    - create a variable 'int progtype' which specifies the plugin type
      FF_SOURCE or FF_EFFECT

    - create an array 'FFParam ff_params[]' for the default parameters
      and null terminate it.
      copy all the necessary values from the 'defaults' array, specify
      minimum, maximum and default values from the xml config file and
      the type of each parameters

    - set the init, deinit, instantiate, deinstantiate functions with
      the appropriate function pointers

    - modify the screenhack function to return after processing each
      frames and rename screenhack to screenhack_frame
      
    - window specific calls have to be removed, because the rendering
      is not done in a window anymore

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <sys/time.h>

#include "screenhackff.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#if HAVE_XSHM_EXTENSION
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif

#include "version.h"

/* #define DEBUG 1 */

char *progname = "xscrff";
XrmDatabase db;
Bool mono_p = False;

static Display *ff_display = NULL;
static int ff_width, ff_height;

static GC ff_pixmap_gc;
static XImage *ff_ximage;
static Pixmap ff_pixmap;
static int ff_display_depth;

static int ff_cc_inited = 0;
static int ff_pfm_index;
static XPixmapFormatValues ff_pfm[3] = {{0,},}; /* supported pixmap formats in 16, 24 and 32bits */
#if HAVE_XSHM_EXTENSION
static int ff_shm;
static XShmSegmentInfo shminfo;
#endif

static int ff_param_count = 0;
static struct timeval basetime;
    
unsigned gettickcount(void)
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return ((tv.tv_sec - basetime.tv_sec)*1000 +
	    (tv.tv_usec - basetime.tv_usec)/1000);
}

static void init_timer(void)
{
    gettimeofday(&basetime, NULL);
}

static void check_colordepth_support(void)
{
    XPixmapFormatValues *xpfm;
    int i, n;
    
    if (ff_display == NULL)
    {
	ff_display = XOpenDisplay(NULL);
	if (ff_display == NULL)
	    return;
    }

    xpfm = XListPixmapFormats(ff_display, &n);

    for (i = 0; i < n; i++)
    {
	if ((ff_pfm[FF_CAP_16BITVIDEO].depth == 0) &&
	    (xpfm[i].depth == 16) && (xpfm[i].bits_per_pixel == 16))
	{
	    memcpy(&(ff_pfm[FF_CAP_16BITVIDEO]), &(xpfm[i]), sizeof(XPixmapFormatValues));
	}
	else
	if ((ff_pfm[FF_CAP_24BITVIDEO].depth == 0) &&
	    (xpfm[i].depth == 24) && (xpfm[i].bits_per_pixel == 24))
	{
	    memcpy(&(ff_pfm[FF_CAP_24BITVIDEO]), &(xpfm[i]), sizeof(XPixmapFormatValues));
	}
	else
	if ((ff_pfm[FF_CAP_32BITVIDEO].depth == 0) &&
	    (((xpfm[i].depth == 32) && (xpfm[i].bits_per_pixel == 32)) ||
	    ((xpfm[i].depth == 24) && (xpfm[i].bits_per_pixel == 32))))
	{
	    memcpy(&(ff_pfm[FF_CAP_32BITVIDEO]), &(xpfm[i]), sizeof(XPixmapFormatValues));
	}
    }
    XFree(xpfm);
    ff_cc_inited = 1;

#ifdef DEBUG
    for (i = 0; i<3; i++)
    {
	printf("%i\tdepth %d bpp %d scanline_pad %d\n", i, ff_pfm[i].depth, ff_pfm[i].bits_per_pixel,
		ff_pfm[i].scanline_pad);
    }
#endif
}

int init_resourcedb(void)
{
    char *defaults_str;
    int i, len;

    for(ff_param_count = 0; ff_params[ff_param_count].name; ff_param_count++)
	;
    
    len = 0;
    for(i = 0; defaults[i]; i++)
	len += strlen(defaults[i]) + 2;
	
    defaults_str = malloc(len);
    if (defaults_str == NULL)
	return 1;

    *defaults_str = 0;
    for(i = 0; defaults[i]; i++)
    {
	if (defaults[i][0] == '.')
	{
	    strcat(defaults_str, "*");
	    strcat(defaults_str, defaults[i]+1);
	}
	else
	{
	    strcat(defaults_str, defaults[i]);
	}
	
	strcat(defaults_str, "\n");
    }
    db = XrmGetStringDatabase(defaults_str);
    
    free(defaults_str);
    return 0;
}

unsigned sh_instantiate(int width, int height, int depth)
{
    int ff_screen_number;
    Screen *ff_screen;

    Window ff_root_window;
    Visual *ff_visual;
#if HAVE_XSHM_EXTENSION
    int ignore, shm_major, shm_minor;
    Bool shm_pixmaps;
#endif
    /* Open a display on the current root window */
    if (ff_display == NULL)
    {
	ff_display = XOpenDisplay(NULL);
	if (ff_display == NULL)
	    return FF_FAIL;
    }

    if (!ff_cc_inited)
	check_colordepth_support();
    
    ff_pfm_index = (depth >> 3) - 2;

#if HAVE_XSHM_EXTENSION
    /* Check for the XShm extension */
    if (XQueryExtension(ff_display, "MIT-SHM", &ignore, &ignore, &ignore))
    {
	if(XShmQueryVersion(ff_display, &shm_major, &shm_minor, &shm_pixmaps) == True)
	{
	    ff_shm = (shm_pixmaps == True) && (XShmPixmapFormat(ff_display) == ZPixmap);
	    
	    #ifdef DEBUG
	    printf("XShm extention version %d.%d %s shared ZPixmaps\n",
		shm_major, shm_minor, ff_shm ? "with" : "without");
	    #endif
	}
	else
	{
	    ff_shm = 0;
	};
    }
#endif
    /* Get the default screen associated with the previously opened display */
    ff_screen_number = DefaultScreen(ff_display);
    ff_screen = DefaultScreenOfDisplay(ff_display);
    ff_visual = DefaultVisual(ff_display, ff_screen_number);

    /* Get the default root window */
    ff_root_window = DefaultRootWindow(ff_display);    
    ff_display_depth = DefaultDepth(ff_display, ff_screen_number);
    
    /* Create an XImage */
    ff_ximage = XCreateImage(ff_display, CopyFromParent, ff_pfm[ff_pfm_index].depth, ZPixmap, 0, NULL,
	      width, height, 32, width * (ff_pfm[ff_pfm_index].bits_per_pixel>>3));
    
    if (ff_ximage == NULL)
    {
	return FF_FAIL;
    }

#if HAVE_XSHM_EXTENSION
    if (ff_shm)
    {
	/* Get the shared memory and check for errors */
	shminfo.shmid = shmget(IPC_PRIVATE, width*height*(depth>>3), IPC_CREAT | 0777);
	if(shminfo.shmid < 0)
	    return FF_FAIL;
	
	/* attach, and check for errrors */
	shminfo.shmaddr = (char *)shmat(shminfo.shmid, 0, 0);
	if (shminfo.shmaddr == (char *) -1)
	    return FF_FAIL;
	
	/* set as read/write, and attach to the display */
	shminfo.readOnly = False;
	if (!XShmAttach(ff_display, &shminfo))
	    return FF_FAIL;

	ff_pixmap = XShmCreatePixmap(ff_display, ff_root_window, shminfo.shmaddr, &shminfo,
	    width, height, ff_pfm[ff_pfm_index].depth);
    }
    else
#endif
    {
	ff_pixmap = XCreatePixmap(ff_display, ff_root_window, width, height, ff_pfm[ff_pfm_index].depth);
    }
    
    ff_pixmap_gc = XCreateGC(ff_display, ff_pixmap, 0, NULL); 

    ff_width = width;
    ff_height = height;
    
    /* This is the one and only place that the random-number generator is
     seeded in any screenhack.  You do not need to seed the RNG again,
     it is done for you before your code is invoked. */
#undef ya_rand_init
    ya_rand_init(0);

    if (init_resourcedb())
	return FF_FAIL;
    
    if (screenhack_instantiate != NULL)
	return screenhack_instantiate(ff_display, ff_pixmap);
    
    return FF_SUCCESS;
}

static PlugInfoStruct  *getInfo(void);
static DWORD		initialise(void);
static DWORD		deInitialise(void);
static DWORD		getNumParameters(void);
static char	       *getParameterName(DWORD i);
static ParameterValue	getParameterDefault(DWORD i);
static DWORD		getPluginCaps(DWORD index);
static PlugExtendedInfoStruct *getExtendedInfo(void);
static DWORD		getParameterType(DWORD i);

static char	       *getParameterDisplay(DWORD i);
static DWORD		setParameter(SetParameterStruct *sps);
static ParameterValue	getParameter(DWORD i);
static DWORD		deInstantiate(void);
static LPVOID		instantiate(VideoInfoStruct *pVideoInfo);
static DWORD		processFrame(LPVOID pFrame);
static DWORD		processFrameCopy(ProcessFrameCopyStruct *pFrameData);

plugMainUnion plugMain(DWORD functionCode, LPVOID pParam, DWORD instanceID)
{
    plugMainUnion retval;
    ParameterValue p;

    switch(functionCode)
    {
	case FF_GETINFO:
	    retval.PISvalue = getInfo();
	    break;
	case FF_INITIALISE:
	    retval.ivalue = initialise();
	    break;
	case FF_DEINITIALISE:
	    retval.ivalue = deInitialise();
	    break;
	case FF_GETNUMPARAMETERS:
	    retval.ivalue = getNumParameters();
	    break;
	case FF_GETPARAMETERNAME:
	    retval.svalue = getParameterName((DWORD)pParam);
	    break;
	case FF_GETPARAMETERDEFAULT:
	    p = getParameterDefault((DWORD)pParam); 
	    retval.ivalue = p.ivalue;
	    break;
	case FF_GETPARAMETERDISPLAY:
	    retval.svalue = getParameterDisplay((DWORD)pParam);
	    break;	
	case FF_SETPARAMETER:
	    retval.ivalue = setParameter((SetParameterStruct *)pParam);
	    break;
	case FF_PROCESSFRAME:
	    retval.ivalue = processFrame(pParam);
	    break;
	case FF_GETPARAMETER:
	    p = getParameter((DWORD)pParam);
	    retval.ivalue = p.ivalue;
	    break;
	case FF_GETPLUGINCAPS:
	    retval.ivalue = getPluginCaps((DWORD)pParam);
	    break;
	case FF_INSTANTIATE:
	    retval.ivalue = (DWORD)instantiate((VideoInfoStruct *)pParam);
	    break;
	case FF_DEINSTANTIATE:
	    retval.ivalue = deInstantiate();
	    break;
	case FF_GETEXTENDEDINFO: 
	    retval.ivalue = (DWORD)getExtendedInfo();
	    break;
	case FF_PROCESSFRAMECOPY:
	    retval.ivalue = processFrameCopy((ProcessFrameCopyStruct*)pParam);
	    break;
	case FF_GETPARAMETERTYPE:
	    retval.ivalue = getParameterType((DWORD)pParam);
	    break;

	default:
	    retval.ivalue = FF_FAIL;
	    break;
    }
    return retval;
}

/*     
 * getInfo
 *
 * gets information about the plugin - version, unique id, short name and type 
 *     
 * input value:    
 *	none
 *     
 * return value:
 *	32-bit pointer to PluginInfoStruct if successful, FF_FAIL otherwise
 */
static PlugInfoStruct *getInfo(void)
{
    static PlugInfoStruct pi;

    if (!ff_cc_inited)
	check_colordepth_support();
    
    pi.APIMajorVersion = 1;
    pi.APIMinorVersion = 000;
    
    strncpy((char *)pi.uniqueID, progclass, 4);
    strncpy((char *)pi.pluginName, progclass, 16);
    pi.pluginType = progtype;

    return &pi;
}

/* 
 * initialise
 *
 * Prepare the plug-in for processing. Set default values, allocate memory. When the
 * plug-in returns from this function it must be ready to accept calls to instantiate.
 *     
 * input value:
 *	none
 *     
 * return value:
 *	FF_SUCCESS if successful, FF_FAIL or other error-codes if failed
 */
static DWORD initialise(void)
{
    init_timer();
    if (screenhack_init != NULL)
	return screenhack_init();
    else
	return FF_SUCCESS;
}

/* 
 * deInitialise
 *
 * tidy up, deallocate memory
 *     
 * input value:
 *	none
 *     
 * return value:
 *	FF_SUCCESS if successful, FF_FAIL or other error-codes if failed
 */
static DWORD deInitialise(void)
{
    if (screenhack_deinit != NULL)
	return screenhack_deinit();
    else
	return FF_SUCCESS;
}

/* 
 * getNumParameters
 *
 * returns number of parameters in plugin
 *     
 * input value:
 *	none
 *
 * return value:
 *	NumParameters indicating the number of parameters in plugin or FF_FAIL on error
 */
static DWORD getNumParameters(void)
{
    return ff_param_count;
}

/*     
 * getParameterName
 *
 * returns pointer to 16 byte char array containing the name of parameter specified by index     
 *
 * input value:
 *	index: ParameterNumber
 *     
 * return value:
 *	32-bit pointer to ParameterName (16 byte char array) containing the name of parameter
 *	specified by index. Returns FF_FAIL on error
 */
static char *getParameterName(DWORD i)
{
    if (i >= ff_param_count)
	return (char *)FF_FAIL;

    return (char *)ff_params[i].name;
}

/*
 * getParameterDefault
 *
 * ParameterDefaultValue indicating the default value of parameter specified by index. For
 * float parameters, the return value is a 32-bit float (0<=value<=1). Return value for text
 * parameters is a 32-bit pointer to a null terminated string.
 *     
 * input value:
 *	index:  ParameterNumber
 *     
 * return value:
 *	ParameterValue or FF_FAIL on error
 */
static ParameterValue getParameterDefault(DWORD i)
{
    ParameterValue p;

    if (i >= ff_param_count)
	p.svalue = (char *)FF_FAIL;
    else /* FIXME: text? */
	p.fvalue = (ff_params[i].def - ff_params[i].min) / 
		    (ff_params[i].max - ff_params[i].min);
    return p;
}

/*     
 * getPluginCaps
 *
 * Indicates capability of feature specified by capsIndex.
 *     
 * input value:
 *	capsIndex: 	PluginCapsIndex
 *     
 * return value:
 *	dword
 *
 */
static DWORD getPluginCaps(DWORD index)
{
    if (!ff_cc_inited)
	check_colordepth_support();
    
    switch (index)
    {
	case FF_CAP_16BITVIDEO:
		return (ff_pfm[FF_CAP_16BITVIDEO].depth != 0) ? FF_SUPPORTED : FF_UNSUPPORTED;
		break;
	case FF_CAP_24BITVIDEO:
    		return (ff_pfm[FF_CAP_24BITVIDEO].depth != 0) ? FF_SUPPORTED : FF_UNSUPPORTED;
		break;
	case FF_CAP_32BITVIDEO: 
		return (ff_pfm[FF_CAP_32BITVIDEO].depth != 0) ? FF_SUPPORTED : FF_UNSUPPORTED;
		break;
	
	case FF_CAP_PROCESSFRAMECOPY:
		return FF_UNSUPPORTED;
		break;
	case FF_CAP_MINIMUMINPUTFRAMES:
		return 1;
		break;
	case FF_CAP_MAXIMUMINPUTFRAMES:
		return 1;
		break;
	case FF_CAP_COPYORINPLACE:
		return FF_CAP_PREFER_NONE;
		break;
	default:
		return FF_UNSUPPORTED;
		break;
    }
}

/* 
 * getExtendedInfo
 * 
 * input value:
 *     none
 * 
 * return value:
 *     pointer to PluginExtendedInfoStruct or FF_FAIL on error
 */
static PlugExtendedInfoStruct *getExtendedInfo(void)
{
    static PlugExtendedInfoStruct pei;
    static char version[256];
    double fver;
    char *s1, *s2, *s3, *s4;    
    char *v;

    v = (char *)strdup(strchr(screensaver_id, ' '));
    s1 = (char *)strchr(v,  ' '); s1++;
    s2 = (char *)strchr(s1, ' ');
    s3 = (char *)strchr(v,  '('); s3++;
    s4 = (char *)strchr(s3, ')');
    *s2 = 0;
    *s4 = 0;
    snprintf(version, 256, "%s: from the XScreenSaver %s distribution (%s.)",
	    progclass, s1, s3);
    fver = strtod(s1, NULL);
    free(v);

    pei.PluginMajorVersion = (int)fver;
    pei.PluginMinorVersion = (int)(100*(fver - pei.PluginMajorVersion));

    pei.Description = version;
    pei.About = "Converted to FreeFrame plugin by Gabor Papp";

    pei.FreeFrameExtendedDataSize = 0;
    pei.FreeFrameExtendedDataBlock = NULL;

    return &pei;
}

/* 
 * getParameterType
 *    
 * input value:
 *     i: ParameterNumber
 *
 * return value:
 *     ParameterType value which tells the host what kind of data the parameter index is or
 *     FF_FAIL on error
 */
static DWORD getParameterType(DWORD i)
{
    if (i >= ff_param_count)
	return FF_FAIL;

    return ff_params[i].type;
}

/* 
 * getParameterDisplay
 * 
 * returns pointer to ParameterDisplayValue containing a string to display the
 * value of the parameter
 *
 * input value:
 *     i: ParameterNumber
 *
 * return value:
 *     pointer or FF_FAIL if failed
 *
 */
static char *getParameterDisplay(DWORD i)
{
    
    if (i >= ff_param_count)
	return (char *)FF_FAIL;

    switch (ff_params[i].type)
    {	
	case FF_TYPE_BOOLEAN:
	case FF_TYPE_EVENT:
	    return get_string_resource(ff_params[i].name, "Boolean");
	    break;

	case FF_TYPE_RED:
	case FF_TYPE_GREEN:
	case FF_TYPE_BLUE:
	case FF_TYPE_XPOS:
	case FF_TYPE_YPOS:
	    return get_string_resource(ff_params[i].name, "Integer");
	    break;
	
	case FF_TYPE_TEXT:
	    return get_string_resource(ff_params[i].name, "String");
	    break;
	    
	default:
	    return get_string_resource(ff_params[i].name, "Float");
	    break;
    }
    
    return (char *)FF_FAIL;
}

/* 
 * setParameter
 * 
 * sets the value of a parameter
 * 
 * input value:
 *     sps: pointer to SetParameterStruct
 * 
 * return value:
 *     returns FF_SUCCESS if successful or FF_FAIL if failed
 */
static DWORD setParameter(SetParameterStruct *sps)
{
    XrmValue xvalue;
    float fvalue;
    char svalue[16];
    
    int i = sps->index;
    if (i >= ff_param_count)
	return FF_FAIL;

    switch (ff_params[i].type)
    {
	case FF_TYPE_BOOLEAN:
	    if (sps->value == 0.0)
	    {
		xvalue.size = 5;
		xvalue.addr = "False";
		XrmPutResource(&db, ff_params[i].name, "Boolean", &xvalue);
	    }
	    else
	    {
		xvalue.size = 4;
		xvalue.addr = "True";
		XrmPutResource(&db, ff_params[i].name, "Boolean", &xvalue);
	    }
	    break;
	case FF_TYPE_STANDARD:
	default:
	    fvalue = sps->value * (ff_params[i].max - ff_params[i].min) + ff_params[i].min;
	    snprintf(svalue, 16, "%16f", fvalue);
	    xvalue.size = 16;
	    xvalue.addr = svalue;
	    XrmPutResource(&db, ff_params[i].name, "Float", &xvalue);
	    break;
    }
	    
    return FF_SUCCESS;
}

/* 
 * getParameter
 * 
 * returns value of a parameter
 * 
 * input value:
 *     index: ParameterNumber
 * 
 * return value:
 *     ParameterValue or FF_FAIL on error
 */
static ParameterValue getParameter(DWORD i)
{
    ParameterValue p;
    p.svalue = (char *)FF_FAIL;

    if (i >= ff_param_count)
	p.svalue = (char *)FF_FAIL;
    else /* FIXME: text */
    {
	switch (ff_params[i].type)
	{
	    case FF_TYPE_BOOLEAN:
		p.fvalue = get_boolean_resource(ff_params[i].name, "Boolean");
		break;
	    case FF_TYPE_STANDARD:
	    default:
		p.fvalue = (get_float_resource(ff_params[i].name, "Float") - ff_params[i].min) / 
		    (ff_params[i].max - ff_params[i].min);
		break;
	}
    }
    return p;
}

static int has_instance = 0;
/* 
 * instantiate
 * 
 * Prepares an instance of the plugin for processing, sets default values, allocates memory.
 * When the plug-in returns from this function it must be ready to process a frame.
 * 
 * input value:
 *     videoInfo: pointer to VideoInfoStruct
 * 
 * return value:
 *     an InstanceIdentifier for the newly instantiated plugin or FF_FAIL on error
 */
static LPVOID instantiate(VideoInfoStruct* pVideoInfo)
{
    int bitconv[] = {16, 24, 32};
    
    /* FIXME - multiple instances */
    if (has_instance)
	return (LPVOID)FF_FAIL;
    
    has_instance = 1;

    sh_instantiate(pVideoInfo->frameWidth, pVideoInfo->frameHeight, bitconv[pVideoInfo->bitDepth]);
    
    /* fake instance id */
    return (LPVOID)0x1234;
}

/* 
 * deInstantiate
 * 
 * closes instance
 * 
 * input value:
 *     none
 * 
 * return value:
 *     FF_SUCCESS or FF_FAIL
 */
static DWORD deInstantiate(void)
{
    XrmDestroyDatabase(db);
    
#if HAVE_XSHM_EXTENSION
    if (ff_shm)
    {
	XShmDetach(ff_display, &shminfo);
	shmdt(shminfo.shmaddr);
	shmctl(shminfo.shmid, IPC_RMID, 0);
    }
#endif
    
    XFreeGC(ff_display, ff_pixmap_gc);
    XFreePixmap(ff_display, ff_pixmap);
    ff_ximage->data = NULL;
    XDestroyImage(ff_ximage);
    
    XCloseDisplay(ff_display);
    
    has_instance = 0;
    
    return FF_SUCCESS;
}

/* 
 * processFrame
 * 
 * Processes a frame of video in place. This is also the basic frameport for source plugins.
 * 
 * input value:
 *     pFrame: pointer to array containing frame of video
 * 
 * return value:
 *     FF_SUCCESS or FF_FAIL if failed
 */
static DWORD processFrame(LPVOID pFrame)
{
    if (progtype == FF_EFFECT)
    {
#if HAVE_XSHM_EXTENSION
	if (ff_shm)
	{
	    memcpy(shminfo.shmaddr, pFrame, ff_width*ff_height*(ff_pfm[ff_pfm_index].bits_per_pixel>>3));
	}
	else
#endif
	{
	    ff_ximage->data = (char *)pFrame;
	    XPutImage(ff_display, ff_pixmap, ff_pixmap_gc, ff_ximage, 0, 0, 0, 0,
		ff_width, ff_height);
	}
    }

    screenhack_frame(ff_display, ff_pixmap);

#if HAVE_XSHM_EXTENSION
    if (ff_shm)
    {
	memcpy(pFrame, shminfo.shmaddr, ff_width*ff_height*(ff_pfm[ff_pfm_index].bits_per_pixel>>3));
    }
    else
#endif
    {
	XImage *xframe;
	xframe = XGetImage(ff_display, ff_pixmap, 0, 0, ff_width, ff_height,
			~0L, ZPixmap);

	memcpy(pFrame, xframe->data, ff_width*ff_height*(ff_pfm[ff_pfm_index].bits_per_pixel>>3));
	XDestroyImage(xframe);
    }

    return FF_SUCCESS;
}

/* 
 * processFrameCopy
 * 
 * performs a source->dest buffer frame process in effects plugins. It is
 * capable of processing multiple input frames. getPluginCaps should be
 * used to see how many input frames the plugin would like.
 * Effect Plugins must support the processFrame (in-place) method of processing.
 * This method is optional.
 * 
 * input value:
 *     pFrameData 32-bit pointer to ProcessFrameCopyStruct
 * 
 * return value:
 *     FF_SUCCESS or FF_FAIL
 */
static DWORD processFrameCopy(ProcessFrameCopyStruct *pFrameData)
{
    return FF_FAIL;
}
